"""
hubspot property groups api
"""
from hubspot3.base import BaseClient
from hubspot3.globals import (
    OBJECT_TYPE_COMPANIES,
    OBJECT_TYPE_CONTACTS,
    OBJECT_TYPE_DEALS,
    OBJECT_TYPE_LINE_ITEMS,
    OBJECT_TYPE_PRODUCTS,
)
from hubspot3.utils import get_log

PROPERTY_GROUPS_API_VERSION = {
    OBJECT_TYPE_COMPANIES: "1",
    OBJECT_TYPE_CONTACTS: "1",
    OBJECT_TYPE_DEALS: "1",
    OBJECT_TYPE_PRODUCTS: "2",
    OBJECT_TYPE_LINE_ITEMS: "2",
}


class PropertyGroupsClient(BaseClient):
    """
    The hubspot3 Property Groups client uses the _make_request method to call the
    API for data.  It returns a python object translated from the json returned
    """

    # Used to store a current object_type. This will be used in `_get_path` to generate the
    # appropriate path to use.
    _object_type = None

    def __init__(self, *args, **kwargs):
        super(PropertyGroupsClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.properties")

    def _get_path(self, subpath):
        return "properties/v{}/{}/groups/{}".format(
            PROPERTY_GROUPS_API_VERSION[self._object_type], self._object_type, subpath
        )

    def create(self, object_type, code, label, extra_params=None):
        self._object_type = object_type

        extra_params = extra_params or {}

        return self._call(
            "", method="POST", data={"name": code, "displayName": label, **extra_params}
        )

    def get_all(self, object_type):
        self._object_type = object_type

        return self._call("", method="GET")

    def delete(self, object_type, code):
        self._object_type = object_type

        return self._call("named/{}".format(code), method="DELETE")

    def delete_all_custom(self, object_type):
        groups_data = self.get_all(object_type)

        for group_data in groups_data:
            group_name = group_data["name"]
            if (
                not group_data["hubspotDefined"]
                # Dirty workaround.
                # Default product properties are currently *not* tagged as `hubspotDefined`... :/
                and not (
                    object_type == "products"
                    and group_name
                    in ("productinformation", "productlineiteminformation")
                )
            ):
                self.delete(object_type, group_name)



"""
base hubspot client class
"""
import http.client
import json
import logging
import time
import traceback
import urllib.request
import urllib.parse
import urllib.error
import zlib
from typing import Callable, List, Optional, Union
from typing_extensions import Literal
from hubspot3 import utils
from hubspot3.utils import force_utf8
from hubspot3.error import (
    HubspotBadConfig,
    HubspotBadRequest,
    HubspotConflict,
    HubspotError,
    HubspotNoConfig,
    HubspotNotFound,
    HubspotRateLimited,
    HubspotServerError,
    HubspotTimeout,
    HubspotUnauthorized,
)


class BaseClient:
    """Base abstract object for interacting with the HubSpot APIs"""

    # These rules are too restrictive for the `__init__()` method and attributes.
    # pylint: disable=too-many-arguments,too-many-instance-attributes

    # Controls how long we sleep for during retries, overridden by unittests
    # so tests run faster
    sleep_multiplier = 1

    def __init__(
        self,
        api_key: str = None,
        access_token: str = None,
        refresh_token: str = None,
        client_id: str = None,
        client_secret: str = None,
        oauth2_token_getter: Optional[
            Callable[[Literal["access_token", "refresh_token"], str], str]
        ] = None,
        oauth2_token_setter: Optional[
            Callable[[Literal["access_token", "refresh_token"], str, str], None]
        ] = None,
        timeout: int = 10,
        mixins: List = None,
        api_base: str = "api.hubapi.com",
        debug: bool = False,
        disable_auth: bool = False,
        **extra_options
    ) -> None:
        super(BaseClient, self).__init__()
        # reverse so that the first one in the list because the first parent
        if not mixins:
            mixins = []
        mixins.reverse()
        for mixin_class in mixins:
            if mixin_class not in self.__class__.__bases__:
                self.__class__.__bases__ = (mixin_class,) + self.__class__.__bases__

        self.api_key = api_key
        # These are used as fallbacks if there aren't setters/getters, or if no remote tokens can be
        # found. The properties without `__` prefixes should generally be used instead of these.
        self.__access_token = access_token
        self.__refresh_token = refresh_token
        self.client_id = client_id
        self.client_secret = client_secret
        if (oauth2_token_getter is None) != (oauth2_token_setter is None):
            raise HubspotBadConfig(
                "You must either specify both the oauth2 token setter and getter, or neither."
            )
        self.oauth2_token_getter = oauth2_token_getter
        self.oauth2_token_setter = oauth2_token_setter
        self.log = utils.get_log("hubspot3")
        if not disable_auth:
            if self.api_key and self.access_token:
                raise HubspotBadConfig("Cannot use both api_key and access_token.")
            if not (self.api_key or self.access_token or self.refresh_token):
                raise HubspotNoConfig("Missing required credentials.")
        self.options = {
            "api_base": api_base,
            "debug": debug,
            "disable_auth": disable_auth,
            "timeout": timeout,
        }
        self.options.update(extra_options)
        self._prepare_connection_type()

    @property
    def credentials(self):
        """
        Credentials to be used when a client needs to instantiate another one.
        Example:
        association_client = AssociationsClient(**self.credentials)
        """
        return {
            "api_key": self.api_key,
            "access_token": self.access_token,
            "refresh_token": self.refresh_token,
            "oauth2_token_getter": self.oauth2_token_getter,
            "oauth2_token_setter": self.oauth2_token_setter,
        }

    @property
    def access_token(self):
        if self.oauth2_token_getter:
            return (
                self.oauth2_token_getter("access_token", self.client_id)
                or self.__access_token
            )
        return self.__access_token

    @access_token.setter
    def access_token(self, access_token):
        if self.oauth2_token_setter:
            self.oauth2_token_setter("access_token", self.client_id, access_token)
        self.__access_token = access_token

    @property
    def refresh_token(self):
        if self.oauth2_token_getter:
            return (
                self.oauth2_token_getter("refresh_token", self.client_id)
                or self.__refresh_token
            )
        return self.__refresh_token

    @refresh_token.setter
    def refresh_token(self, refresh_token):
        if self.oauth2_token_setter:
            self.oauth2_token_setter("refresh_token", self.client_id, refresh_token)
        else:
            self.__refresh_token = refresh_token

    def _prepare_connection_type(self):
        connection_types = {
            "http": http.client.HTTPConnection,
            "https": http.client.HTTPSConnection,
        }
        parts = self.options["api_base"].split("://")
        protocol = (parts[0:-1] + ["https"])[0]
        self.options["connection_type"] = connection_types[protocol]
        self.options["protocol"] = protocol
        self.options["api_base"] = parts[-1]

    def _get_path(self, subpath):
        """get the full api url for the given subpath on this client"""
        return subpath

    def _prepare_request_auth(self, subpath, params, data, opts):
        if self.api_key:
            params["hapikey"] = params.get("hapikey") or self.api_key

    def _prepare_request(
        self,
        subpath,
        params,
        data,
        opts,
        doseq=False,
        query="",
        retried=False,
        properties=None,
    ):
        params = params or {}
        properties = properties or []
        self._prepare_request_auth(subpath, params, data, opts)

        if opts.get("hub_id") or opts.get("portal_id"):
            params["portalId"] = opts.get("hub_id") or opts.get("portal_id")
        if query is None:
            query = ""
        if query and query.startswith("?"):
            query = query[1:]
        if query and not query.startswith("&"):
            query = "&" + query
        url = opts.get("url") or "/{}?{}{}".format(
            self._get_path(subpath), urllib.parse.urlencode(params, doseq), query
        )
        headers = opts.get("headers") or {}
        headers.update(
            {
                "Accept-Encoding": "gzip",
                "Content-Type": opts.get("content_type") or "application/json",
            }
        )
        if self.access_token:
            headers.update({"Authorization": "Bearer {}".format(self.access_token)})

        if data and headers["Content-Type"] == "application/json" and not retried:
            data = json.dumps(data)

        for hs_property in properties:
            url += "&properties={}".format(hs_property)

        return url, headers, data

    def _create_request(self, conn, method, url, headers, data):
        conn.request(method, url, data, headers)
        params = {
            "method": method,
            "url": url,
            "data": data,
            "headers": headers,
            "host": conn.host,
            "timeout": conn.timeout,
        }
        return params

    def _gunzip_body(self, body):
        return zlib.decompress(body)

    def _process_body(self, data, gzipped):
        if gzipped:
            return force_utf8(self._gunzip_body(data))
        return data

    def _execute_request_raw(self, conn, request):
        try:
            result = conn.getresponse()
        except Exception:
            raise HubspotTimeout(None, request, traceback.format_exc())

        encoding = [i[1] for i in result.getheaders() if i[0] == "content-encoding"]
        possibly_encoded = result.read()
        try:
            possibly_encoded = zlib.decompress(possibly_encoded, 16 + zlib.MAX_WBITS)
        except Exception:
            pass
        result.body = self._process_body(
            possibly_encoded, len(encoding) and encoding[0] == "gzip"
        )

        conn.close()
        if result.status in (404, 410):
            raise HubspotNotFound(result, request)
        if result.status == 401:
            raise HubspotUnauthorized(result, request)
        if result.status == 409:
            raise HubspotConflict(result, request)
        if result.status == 429:
            raise HubspotRateLimited(result, request)
        if 400 <= result.status < 500 or result.status == 501:
            raise HubspotBadRequest(result, request)
        if result.status >= 500:
            raise HubspotServerError(result, request)

        return result

    def _execute_request(self, conn, request):
        result = self._execute_request_raw(conn, request)
        return result.body

    def _digest_result(self, data):
        if data:
            if isinstance(data, bytes):
                data = utils.force_utf8(data)
            if isinstance(data, str):
                try:
                    data = json.loads(data)
                except ValueError:
                    pass

        return data

    def _prepare_request_retry(self, method, url, headers, data):
        pass

    def _call_raw(
        self,
        subpath,
        params=None,
        method="GET",
        data=None,
        doseq=False,
        query="",
        retried=False,
        properties=None,
        **options
    ):
        opts = self.options.copy()
        opts.update(options)

        debug = opts.get("debug")

        url, headers, data = self._prepare_request(
            subpath,
            params,
            data,
            opts,
            doseq=doseq,
            query=query,
            retried=retried,
            properties=properties,
        )

        if debug:
            print(
                json.dumps(
                    {"url": url, "headers": headers, "data": data},
                    sort_keys=True,
                    indent=2,
                )
            )

        kwargs = {"timeout": opts["timeout"]}

        num_retries = opts.get("number_retries", 2)

        # Never retry a POST, PUT, or DELETE unless explicitly told to
        if method != "GET" and not opts.get("retry_on_post"):
            num_retries = 0
        if num_retries > 6:
            num_retries = 6

        emergency_brake = 10
        try_count = 0
        while True:
            emergency_brake -= 1
            # avoid getting burned by any mistakes in While loop logic
            if emergency_brake < 1:
                break
            try:
                try_count += 1
                connection = opts["connection_type"](opts["api_base"], **kwargs)
                request_info = self._create_request(
                    connection, method, url, headers, data
                )
                result = self._execute_request_raw(connection, request_info)
                break
            except HubspotUnauthorized:
                self.log.warning("401 Unauthorized response to API request.")
                if (
                    self.access_token
                    and self.refresh_token
                    and self.client_id
                    and self.client_secret
                ):
                    if retried:
                        self.log.error(
                            "Refreshed token, but request still was not authorized. "
                            "You may need to grant additional permissions."
                        )
                        raise

                    from hubspot3.oauth2 import OAuth2Client

                    self.log.info("Refreshing access token")
                    try:
                        client = OAuth2Client(**self.options)
                        refresh_result = client.refresh_tokens(
                            client_id=self.client_id,
                            client_secret=self.client_secret,
                            refresh_token=self.refresh_token,
                        )
                        self.access_token = refresh_result["access_token"]
                        self.refresh_token = refresh_result["refresh_token"]
                        self.log.info(
                            "Retrying with new token {}".format(self.access_token)
                        )
                    except Exception as exception:
                        self.log.error(
                            "Unable to refresh access_token: {}".format(exception)
                        )
                        raise
                    return self._call_raw(
                        subpath,
                        params=params,
                        method=method,
                        data=data,
                        doseq=doseq,
                        query=query,
                        retried=True,
                        **options
                    )
                if self.access_token:
                    self.log.warning(
                        "In order to enable automated refreshing of your access token, please "
                        "provide a client ID, client secret and refresh token in addition to the "
                        "access token."
                    )
                raise
            except HubspotError as exception:
                if try_count > num_retries:
                    logging.warning("Too many retries for {}".format(url))
                    raise
                # Don't retry errors from 300 to 499
                if exception.result and 300 <= exception.result.status < 500:
                    raise
                self._prepare_request_retry(method, url, headers, data)
                self.log.warning(
                    "HubspotError {} calling {}, retrying".format(exception, url)
                )
            # exponential back off
            # wait 0 seconds, 1 second, 3 seconds, 7 seconds, 15 seconds, etc
            time.sleep((pow(2, try_count - 1) - 1) * self.sleep_multiplier)
        return result

    def _call(
        self,
        subpath: str,
        params: dict = None,
        method: str = "GET",
        data: Union[str, dict, list] = None,
        doseq: bool = False,
        query: str = "",
        raw: bool = False,
        properties: list = None,
        **options
    ):
        result = self._call_raw(
            subpath,
            params=params,
            method=method,
            data=data,
            doseq=doseq,
            query=query,
            retried=False,
            properties=properties,
            **options
        )
        return result if raw else self._digest_result(result.body)



"""
hubspot broadcast api
"""
from hubspot3.base import BaseClient


HUBSPOT_BROADCAST_API_VERSION = "1"


class BaseSocialObject:
    """base social object"""

    def _camel_case_to_underscores(self, text):
        result = []
        pos = 0
        while pos < len(text):
            if text[pos].isupper():
                if (
                    pos - 1 > 0
                    and text[pos - 1].islower()
                    or pos - 1 > 0
                    and pos + 1 < len(text)
                    and text[pos + 1].islower()
                ):
                    result.append("_{}".format(text[pos].lower()))
                else:
                    result.append(text[pos].lower())
            else:
                result.append(text[pos])
            pos += 1
        return "".join(result)

    def _underscores_to_camel_case(self, text):
        result = []
        pos = 0
        while pos < len(text):
            if text[pos] == "_" and pos + 1 < len(text):
                result.append("{}".format(text[pos + 1].upper()))
                pos += 1
            else:
                result.append(text[pos])
            pos += 1
        return "".join(result)

    def to_dict(self):
        dict_self = {}
        for key in vars(self):
            dict_self[self._underscores_to_camel_case(key)] = getattr(self, key)
        return dict_self

    def from_dict(self, data):
        accepted_fields = self.accepted_fields()
        for key in data:
            if key in accepted_fields:
                setattr(self, self._camel_case_to_underscores(key), data[key])


class Broadcast(BaseSocialObject):
    """Defines a social media broadcast message for the broadcast api"""

    # Constants for remote content type
    COS_LP = "coslp"
    COS_BLOG = "cosblog"
    LEGACY_LP = "cmslp"
    LEGACY_BLOG = "cmsblog"

    def __init__(self, broadcast_data):
        self.data_parse(broadcast_data)

    def accepted_fields(self):
        return [
            "broadcastGuid",
            "campaignGuid",
            "channel",
            "channelGuid",
            "clicks",
            "clientTag",
            "content",
            "createdAt",
            "createdBy",
            "finishedAt",
            "groupGuid",
            "interactions",
            "interactionCounts",
            "linkGuid",
            "message",
            "messageUrl",
            "portalId",
            "remoteContentId",
            "remoteContentType",
            "status",
            "triggerAt",
            "updatedBy",
        ]

    def data_parse(self, broadcast_data):
        self.from_dict(broadcast_data)


class Channel(BaseSocialObject):
    """Defines the social media channel for the broadcast api"""

    def __init__(self, channel_data):
        self.data_parse(channel_data)

    def accepted_fields(self):
        return [
            "channelGuid",
            "accountGuid",
            "account",
            "type",
            "name",
            "dataMap",
            "createdAt",
            "settings",
        ]

    def data_parse(self, channel_data):
        self.from_dict(channel_data)


class BroadcastClient(BaseClient):
    """Broadcast API to manage messages published to social networks"""

    def _get_path(self, subpath):
        return "broadcast/v{}/{}".format(HUBSPOT_BROADCAST_API_VERSION, subpath)

    def get_broadcast(self, broadcast_guid, **kwargs):
        """
        Get a specific broadcast by guid
        """
        params = kwargs
        broadcast = self._call(
            "broadcasts/{}".format(broadcast_guid),
            params=params,
            content_type="application/json",
        )
        return Broadcast(broadcast)

    def get_broadcasts(
        self, broadcast_type="", page=None, remote_content_id=None, limit=None, **kwargs
    ):
        """
        Get all broadcasts, with optional paging and limits.
        Type filter can be 'scheduled', 'published' or 'failed'
        """
        if remote_content_id:
            return self.get_broadcasts_by_remote(remote_content_id)

        params = {"type": broadcast_type}
        if page:
            params["page"] = page

        params.update(kwargs)

        result = self._call(
            "broadcasts", params=params, content_type="application/json"
        )
        broadcasts = [Broadcast(b) for b in result]

        if limit:
            return broadcasts[:limit]
        return broadcasts

    def create_broadcast(self, broadcast):
        if not isinstance(broadcast, dict):
            return self._call(
                "broadcasts",
                data=broadcast.to_dict(),
                method="POST",
                content_type="application/json",
            )
        return self._call(
            "broadcasts", data=broadcast, method="POST", content_type="application/json"
        )

    def cancel_broadcast(self, broadcast_guid):
        """
        Cancel a broadcast specified by guid
        """
        subpath = "broadcasts/{}/update".format(broadcast_guid)
        broadcast = {"status": "CANCELED"}
        bcast_dict = self._call(
            subpath, method="POST", data=broadcast, content_type="application/json"
        )
        return bcast_dict

    def get_channel(self, channel_guid):
        channel = self._call(
            "channels/{}".format(channel_guid), content_type="application/json"
        )
        return Channel(channel)

    def get_channels(self, current=True, publish_only=False, settings=False):
        """
        if "current" is false it will return all channels that a user
        has published to in the past.

        if publish_only is set to true, then return only the channels
        that are publishable.

        if settings is true, the API will make extra queries to return
        the settings for each channel.
        """
        if publish_only:
            if current:
                endpoint = "channels/setting/publish/current"
            else:
                endpoint = "channels/setting/publish"
        else:
            if current:
                endpoint = "channels/current"
            else:
                endpoint = "channels"

        result = self._call(
            endpoint, content_type="application/json", params=dict(settings=settings)
        )
        return [Channel(c) for c in result]



"""
hubspot workflows api
"""
from hubspot3.base import BaseClient
from hubspot3.utils import get_log

WORKFLOWS_API_VERSION = "3"


class WorkflowsClient(BaseClient):
    """
    The hubspot3 Workflows client uses the _make_request method to call the
    API for data.  It returns a python object translated from the json returned
    """

    def __init__(self, *args, **kwargs):
        """initialize a workflows client"""
        super(WorkflowsClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.workflows")

    def _get_path(self, subpath):
        return "automation/v{}/{}".format(WORKFLOWS_API_VERSION, subpath)

    def get_all_workflow_ids(self, **options):
        """
        Get all workflow IDs
        :see: https://developers.hubspot.com/docs/methods/workflows/v3/get_workflows
        """
        return self._call("workflows", **options)

    def get_workflow_by_id(self, workflow_id: int = None, **options):
        """
        Get workflow specified by ID
        :see: https://developers.hubspot.com/docs/methods/workflows/v3/get_workflow
        """
        if workflow_id is not None:
            return self._call("workflows/{}".format(workflow_id))
        return None



"""
globals file for hubspot3
"""
__version__ = "3.2.50"


BASE_URL = "https://api.hubapi.com"


OBJECT_TYPE_COMPANIES = "companies"
OBJECT_TYPE_CONTACTS = "contacts"
OBJECT_TYPE_DEALS = "deals"
OBJECT_TYPE_PRODUCTS = "products"
OBJECT_TYPE_LINE_ITEMS = "line_items"

DATA_TYPE_BOOL = "bool"
DATA_TYPE_DATETIME = "datetime"
DATA_TYPE_DATE = "date"
DATA_TYPE_ENUM = "enumeration"
DATA_TYPE_NUMBER = "number"
DATA_TYPE_STRING = "string"

VALID_PROPERTY_DATA_TYPES = (
    DATA_TYPE_BOOL,
    DATA_TYPE_DATETIME,
    DATA_TYPE_DATE,
    DATA_TYPE_ENUM,
    DATA_TYPE_NUMBER,
    DATA_TYPE_STRING,
)

WIDGET_TYPE_BOOLEAN_CHECKBOX = "booleancheckbox"
WIDGET_TYPE_CHECKBOX = "checkbox"
WIDGET_TYPE_DATE = "date"
WIDGET_TYPE_FILE = "file"
WIDGET_TYPE_NUMBER = "number"
WIDGET_TYPE_RADIO = "radio"
WIDGET_TYPE_SELECT = "select"
WIDGET_TYPE_TEXT = "text"
WIDGET_TYPE_TEXTAREA = "textarea"

VALID_PROPERTY_WIDGET_TYPES = (
    WIDGET_TYPE_BOOLEAN_CHECKBOX,
    WIDGET_TYPE_CHECKBOX,
    WIDGET_TYPE_DATE,
    WIDGET_TYPE_FILE,
    WIDGET_TYPE_NUMBER,
    WIDGET_TYPE_RADIO,
    WIDGET_TYPE_SELECT,
    WIDGET_TYPE_TEXT,
    WIDGET_TYPE_TEXTAREA,
)



"""
hubspot forms api
"""
import json
from urllib.parse import urlencode
from http.client import HTTPResponse
from hubspot3.base import BaseClient
from hubspot3.error import HubspotNotFound, HubspotServerError
from typing import Dict


FORMS_API_VERSION = 2


class FormSubmissionClient(BaseClient):
    """allows acccess to submit forms via the hubspot api"""

    class ResponseCode:
        """form submission response codes enum"""

        SUCCESS = 204
        SUCCESS_AND_REDIRECT = 302
        NOT_FOUND = 404
        ERROR = 500

    def __init__(self, *args, **kwargs) -> None:
        """form submission client constructor"""
        kwargs.update({"disable_auth": True})
        super(FormSubmissionClient, self).__init__(*args, **kwargs)
        self.options["api_base"] = "forms.hubspot.com"

    def _get_path(self, subpath: str) -> str:
        """api path for submitting to a form"""
        return "/uploads/form/v{}/{}".format(FORMS_API_VERSION, subpath)

    def submit_form(
        self,
        portal_id: str,
        form_guid: str,
        data: dict,
        context: dict = None,
        **options
    ) -> HTTPResponse:
        """
        submit to a form on hubspot
        does not require credentials

        this will urlencode the data automatically
        and if you pass a context:
            it will json dump for you, and place inside the data

        :see: https://developers.hubspot.com/docs/methods/forms/submit_form
        """
        if context:
            data["hs_context"] = json.dumps(context)
        subpath = "{}/{}".format(portal_id, form_guid)
        opts = {"content_type": "application/x-www-form-urlencoded"}
        options.update(opts)
        response = self._call(
            subpath, method="POST", data=urlencode(data), raw=True, **options
        )
        if response.status in [
            FormSubmissionClient.ResponseCode.SUCCESS,
            FormSubmissionClient.ResponseCode.SUCCESS_AND_REDIRECT,
        ]:
            return response
        if response.status == FormSubmissionClient.ResponseCode.NOT_FOUND:
            raise HubspotNotFound(response, None)
        if response.status == FormSubmissionClient.ResponseCode.ERROR:
            raise HubspotServerError(response, None)

        # shouldn't ever get here, but raise anyways
        raise HubspotServerError(response, None)


class FormsClient(BaseClient):
    """allows access to form data inside your hubspot account"""

    def __init__(self, *args, **kwargs) -> None:
        """forms client constructor"""
        super(FormsClient, self).__init__(*args, **kwargs)

    def _get_path(self, subpath) -> str:
        """api path for the forms related calls"""
        return "forms/v{}/{}".format(FORMS_API_VERSION, subpath)

    def get(self, form_id: str, **options) -> Dict:
        """
        get a form by its form_id
        :see: https://developers.hubspot.com/docs/methods/forms/v2/get_form
        """
        return self._call("forms/{}".format(form_id), method="GET", **options)

    def get_all(self, limit: int = -1, offset: int = 0, **options) -> list:
        """
        get all forms from this hubspot portal.
        :see: https://developers.hubspot.com/docs/methods/forms/v2/get_forms
        """
        params = {"offset": offset}
        if limit > 0:
            params["limit"] = limit
        return self._call("forms", method="GET", params=params)



"""
hubspot companies properties api
"""
from hubspot3.base import BaseClient
from hubspot3.utils import get_log
from typing import List, Optional, Union


COMPANIES_PROPERTIES_API_VERSION = "1"


class CompaniesPropertiesClient(BaseClient):
    """
    The hubspot3 Companies Properties client uses the _make_request method to call the
    API for data.  It returns a python object translated from the json returned.
    """

    def __init__(self, *args, **kwargs):
        """initialize an companies properties client"""
        super(CompaniesPropertiesClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.companies_properties")

    def _get_path(self, subpath: str) -> str:
        return "properties/v{}/companies/{}".format(
            COMPANIES_PROPERTIES_API_VERSION, subpath
        )

    def get_all_companies_properties(
        self, extra_properties: Union[str, List] = None, **options
    ) -> Optional[List]:
        """
        Retrieve all of the company properties, including their definition, for a given account.
        :see: https://developers.hubspot.com/docs/methods/companies/get_company_properties
        """
        return self._call("properties", **options)



"""
hubspot3 error helpers
"""
import json
from urllib import parse

from hubspot3.utils import force_utf8


class EmptyResult:
    """
    Null Object pattern to prevent Null reference errors
    when there is no result
    """

    def __init__(self):
        self.status = 0
        self.body = ""
        self.msg = ""
        self.reason = ""


class HubspotError(ValueError):
    """Any problems get thrown as HubspotError exceptions with the relevant info inside"""

    as_str_template = """
{error_message}

---- request ----
{method} {host}{url}, [timeout={timeout}]

---- body ----
{body}

---- headers ----
{headers}

---- result ----
{result_status}

---- body -----
{result_body}

---- headers -----
{result_headers}

---- reason ----
{result_reason}

---- trigger error ----
{error}
        """

    def __contains__(self, item):
        """tests if the given item text is in the error text"""
        return item in self.__str__()

    def __init__(self, result, request, err=None):
        super(HubspotError, self).__init__(result and result.reason or "Unknown Reason")
        if result is None:
            self.result = EmptyResult()
        else:
            self.result = result
        if request is None:
            request = {}
        if "url" in request:
            request["url"] = self._uglify_hapikey(request["url"])
        self.request = request
        self.err = err

    def __str__(self):
        params = {}
        request_keys = ("method", "host", "url", "data", "headers", "timeout", "body")
        result_attrs = ("status", "reason", "msg", "body", "headers")
        params["error"] = self.err
        params["error_message"] = "Hubspot Error"
        if self.result.body:
            try:
                result_body = json.loads(force_utf8(self.result.body))
            except ValueError:
                result_body = {}
            params["error_message"] = result_body.get("message", "Hubspot Error")
        for key in request_keys:
            params[key] = self.request.get(key)
        for attr in result_attrs:
            params["result_{}".format(attr)] = getattr(self.result, attr, "")

        params = self._dict_vals_to_str(params)
        return self.as_str_template.format(**params)

    def _dict_vals_to_str(self, data):
        unicode_data = {}
        for key, val in data.items():
            if not val:
                unicode_data[key] = ""
            if isinstance(val, bytes):
                unicode_data[key] = force_utf8(val)
            elif isinstance(val, str):
                unicode_data[key] = val
            else:
                unicode_data[key] = str(type(val))
        return unicode_data

    def _uglify_hapikey(self, url):
        url_parse = parse.urlparse(url)
        parse_query = parse.parse_qs(url_parse.query)
        if "hapikey" not in parse_query:
            return url
        parse_query["hapikey"][0] = "{}****".format(parse_query["hapikey"][0][0:4])
        parse_result = parse.ParseResult(
            scheme=url_parse.scheme,
            netloc=url_parse.netloc,
            path=url_parse.path,
            params=url_parse.params,
            query=parse.urlencode(parse_query, doseq=True, safe="*"),
            fragment=url_parse.fragment,
        )
        return parse.urlunparse(parse_result)


class HubspotBadConfig(Exception):
    """misconfigured api_key and/or access_token credentials client"""


class HubspotNoConfig(Exception):
    """no api_key or access_token credentials were passed to the client"""


# Create more specific error cases, to make filtering errors easier
class HubspotBadRequest(HubspotError):
    """most 40X results and 501 results"""


class HubspotNotFound(HubspotError):
    """404 and 410 results"""


class HubspotTimeout(HubspotError):
    """socket timeouts, sslerror, and 504"""


class HubspotUnauthorized(HubspotError):
    """401 Unauthorized errors"""


class HubspotConflict(HubspotError):
    """409 conflict errors"""


class HubspotServerError(HubspotError):
    """most 500 errors"""


class HubspotRateLimited(HubspotError):
    """exception for when we're rate limited"""



"""
hubspot OAuth2 api
"""
from urllib.parse import urlencode
from hubspot3.base import BaseClient
from hubspot3.utils import get_log


OAUTH2_API_VERSION = "1"


class OAuth2Client(BaseClient):
    """
    The hubspot3 OAuth2 client uses the _make_request method to call the
    API for data.  It returns a python object translated from the json returned
    """

    def __init__(self, *args, **kwargs):
        """initialize a contacts client"""
        # Since this client is used to generate tokens for authentication, it does not require
        # authentication itself.
        kwargs["disable_auth"] = True
        super(OAuth2Client, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.oauth2")
        self.options["content_type"] = "application/x-www-form-urlencoded"
        # Make sure that certain credentials that wouldn't be used anyway are not set. Not having
        # an access token will also make sure that the `_call_raw` implementation does not try to
        # refresh access tokens on its own.
        self.api_key = None
        self.access_token = None

    def _get_path(self, subpath: str) -> str:
        return "oauth/v{}/{}".format(OAUTH2_API_VERSION, subpath)

    def get_tokens(
        self,
        authorization_code: str,
        redirect_uri: str,
        client_id: str = None,
        client_secret: str = None,
        **options
    ):
        """
        Request an initial token pair using the provided credentials.

        If any of the optional parameters are not provided, their value will be read from the
        corresponding attributes on this client.
        If the value for all optional parameters had to be read from the attributes, the refresh
        token returned from the API will be stored on this client to allow for further
        `refresh_token` calls without having to provide the refresh token.

        :see: https://developers.hubspot.com/docs/methods/oauth2/get-access-and-refresh-tokens
        """
        data = {
            "grant_type": "authorization_code",
            "client_id": client_id or self.client_id,
            "client_secret": client_secret or self.client_secret,
            "redirect_uri": redirect_uri,
            "code": authorization_code,
        }
        result = self._call("token", method="POST", data=urlencode(data), **options)

        if not client_id and not client_secret:
            self.refresh_token = result["refresh_token"]
        return result

    def refresh_tokens(
        self,
        client_id: str = None,
        client_secret: str = None,
        refresh_token: str = None,
        **options
    ):
        """
        Request a new token pair using the provided refresh token and credentials.

        If any of the optional parameters are not provided, their value will be read from the
        corresponding attributes on this client.
        If the value for all optional parameters had to be read from the attributes, the refresh
        token returned from the API will be stored on this client to allow for further
        `refresh_token` calls without having to provide the refresh token.

        :see: https://developers.hubspot.com/docs/methods/oauth2/refresh-access-token
        """
        data = {
            "grant_type": "refresh_token",
            "client_id": client_id or self.client_id,
            "client_secret": client_secret or self.client_secret,
            "refresh_token": refresh_token or self.refresh_token,
        }
        result = self._call("token", method="POST", data=urlencode(data), **options)

        if not client_id and not client_secret and not refresh_token:
            self.refresh_token = result["refresh_token"]
        return result

    def get_access_token_data(self, access_token: str, **options):
        """
        Get the meta data for an access token.

        :see: https://developers.hubspot.com/docs/methods/oauth2/get-access-token-information
        """
        return self._call("access-tokens/{}".format(access_token), **options)

    def get_refresh_token_data(self, refresh_token: str = None, **options):
        """
        Get the meta data for a refresh token.

        If any of the optional parameters are not provided, their value will be read from the
        corresponding attributes on this client.

        :see: https://developers.hubspot.com/docs/methods/oauth2/get-refresh-token-information
        """
        return self._call(
            "refresh-tokens/{}".format(refresh_token or self.refresh_token), **options
        )

    def delete_refresh_token(self, refresh_token: str = None):
        """
        Deletes a refresh token. You can use this to delete your refresh token if a user
        uninstalls your app.

        If any of the optional parameters are not provided, their value will be read from the
        corresponding attributes on this client.

        :see: https://developers.hubspot.com/docs/methods/oauth2/delete-refresh-token
        """
        return self._call(
            "refresh-tokens/{}".format(refresh_token or self.refresh_token),
            method="DELETE",
        )



"""
hubspot cms_files api
"""
from hubspot3.base import BaseClient


CMS_FILES_API_VERSION = "2"


class CMSFilesClient(BaseClient):
    """
    provides a client for accessing cms files
    """

    def _get_path(self, subpath: str) -> str:
        return "filemanager/api/v{}/files/{}".format(CMS_FILES_API_VERSION, subpath)

    def get_file_meta_data(self, file_id: int, **options):
        return self._call("{}".format(file_id), **options)



"""
hubspot properties api
"""
from typing import Dict, Optional
from hubspot3.base import BaseClient
from hubspot3.globals import (
    OBJECT_TYPE_COMPANIES,
    OBJECT_TYPE_CONTACTS,
    OBJECT_TYPE_DEALS,
    OBJECT_TYPE_LINE_ITEMS,
    OBJECT_TYPE_PRODUCTS,
    VALID_PROPERTY_DATA_TYPES,
    VALID_PROPERTY_WIDGET_TYPES,
    DATA_TYPE_ENUM,
)
from hubspot3.utils import get_log


PROPERTIES_API_VERSION = {
    OBJECT_TYPE_COMPANIES: "1",
    OBJECT_TYPE_CONTACTS: "1",
    OBJECT_TYPE_DEALS: "1",
    OBJECT_TYPE_LINE_ITEMS: "2",
    OBJECT_TYPE_PRODUCTS: "2",
}


class PropertiesClient(BaseClient):
    """
    The hubspot3 Properties client uses the _make_request method to call the
    API for data.  It returns a python object translated from the json returned
    """

    # Used to store a current object_type. This will be used in `_get_path` to generate the
    # appropriate path to use.
    _object_type = None

    def __init__(self, *args, **kwargs):
        super(PropertiesClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.properties")

    def _get_path(self, subpath):
        return "properties/v{}/{}/properties/{}".format(
            PROPERTIES_API_VERSION[self._object_type], self._object_type, subpath
        )

    @staticmethod
    def _validate(
        data_type: Optional[str], widget_type: Optional[str], extra_params: dict
    ) -> None:
        if data_type:
            if data_type not in VALID_PROPERTY_DATA_TYPES:
                raise ValueError(
                    "Invalid data type for property. Valid data types are: {}".format(
                        VALID_PROPERTY_DATA_TYPES
                    )
                )

            if data_type == DATA_TYPE_ENUM and (
                not extra_params or not extra_params.get("options")
            ):
                raise ValueError(
                    "Invalid data for updating an enumeration type. "
                    "Its 'options' have to be specified."
                )

        if widget_type and widget_type not in VALID_PROPERTY_WIDGET_TYPES:
            raise ValueError(
                "Invalid widget type for property. Valid widget types are: {}".format(
                    VALID_PROPERTY_WIDGET_TYPES
                )
            )

    def create(
        self,
        object_type: str,
        code: str,
        label: str,
        description: str,
        group_code: str,
        data_type: str,
        widget_type: str,
        extra_params: dict = None,
    ) -> Dict:
        """
        Create a new custom property on hubspot.
        """
        extra_params = extra_params or {}

        self._validate(data_type, widget_type, extra_params)

        # Save the current object type.
        self._object_type = object_type

        return self._call(
            "",
            method="POST",
            data={
                "name": code,
                "label": label,
                "description": description,
                "groupName": group_code,
                "type": data_type,
                "fieldType": widget_type,
                **extra_params,
            },
        )

    def update(
        self,
        object_type: str,
        code: str,
        label: str = None,
        description: str = None,
        group_code: str = None,
        data_type: str = None,
        widget_type: str = None,
        extra_params: dict = None,
    ) -> Dict:
        """
        Update a custom property on hubspot.
        """

        extra_params = extra_params or {}

        self._validate(data_type, widget_type, extra_params)

        # Save the current object type.
        self._object_type = object_type

        fields = {
            "label": label,
            "description": description,
            "groupName": group_code,
            "type": data_type,
            "fieldType": widget_type,
            **extra_params,
        }
        data = {key: value for key, value in fields.items() if value is not None}

        return self._call("named/{}".format(code), method="PUT", data=data)

    def get_all(self, object_type):
        """Retrieve all the custom properties."""

        # Save the current object type.
        self._object_type = object_type

        return self._call(
            "", method="GET", params={"properties": ["name", "label", "description"]}
        )

    def get(self, object_type: str, code: str):
        """Retrieve a property."""

        # Save the current object type.
        self._object_type = object_type

        return self._call("named/{}".format(code), method="GET")

    def delete(self, object_type, code):
        """Delete a custom property."""

        # Save the current object type.
        self._object_type = object_type

        return self._call("named/{}".format(code), method="DELETE")

    def delete_all(self, object_type):
        """Delete all the custom properties. Please use it carefully."""
        props_data = self.get_all(object_type)

        for prop_data in props_data:
            if not prop_data["hubspotDefined"]:
                self.delete(object_type, prop_data["name"])



"""
base utils for the hubspot3 library
"""
import logging
import sys
from collections import OrderedDict
from typing import Dict, Union


PY_VERSION = sys.version_info


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


def get_log(name: str):
    logger = logging.getLogger(name)
    logger.addHandler(NullHandler())
    return logger


def force_utf8(raw):
    """Will force the string to convert to valid utf8"""
    string = raw
    try:
        string = string.decode("utf-8", "ignore")
    except Exception:
        pass
    return string


def prettify(obj_with_props, id_key):
    prettified = {
        prop: obj_with_props["properties"][prop]["value"]
        for prop in obj_with_props["properties"]
    }
    prettified["id"] = obj_with_props[id_key]
    try:
        prettified.update(
            {
                assoc: obj_with_props["associations"][assoc]
                for assoc in obj_with_props["associations"]
            }
        )
    except KeyError:
        pass

    return prettified


def ordered_dict(dictionary: Dict) -> Union[Dict, OrderedDict]:
    """
    shim to fix some param ordering issues with python 3.5
    in python 3.6+, dictionaries are ordered by default
    """
    if PY_VERSION[0] == 3 and PY_VERSION[1] == 5:
        return OrderedDict(sorted(dictionary.items(), key=lambda x: x[0]))
    return dictionary



"""
hubspot email events api
"""
from hubspot3.base import BaseClient
from hubspot3.utils import get_log


EMAIL_EVENTS_API_VERSION = "1"


class EmailEventsClient(BaseClient):
    """
    The hubspot3 Email Events client uses the _make_request method to call the
    API for data.  It returns a python object translated from the json returned
    """

    def __init__(self, *args, **kwargs):
        """initialize an email events client"""
        super(EmailEventsClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.email_events")

    def _get_path(self, subpath):
        return "email/public/v{}/campaigns/{}".format(EMAIL_EVENTS_API_VERSION, subpath)

    def get_all_campaigns_ids(self, **options):
        """
        Retrieve all email campaign IDs associated with the portal.
        :see: https://developers.hubspot.com/docs/methods/email/get_campaigns_by_id
        """
        return self._call("by-id", **options)

    def get_campaign_data(self, campaign_id: int = None, **options):
        """
        Retrieve campaign data for a given campaign ID.
        :see: https://developers.hubspot.com/docs/methods/email/get_campaign_data
        :param campaign_id:
        """
        if campaign_id is not None:
            return self._call(str(campaign_id))
        return None



"""
hubspot settings api
"""
from hubspot3.base import BaseClient
from hubspot3.error import HubspotError
from typing import Dict


SETTINGS_API_VERSION = "v1"


class SettingsClient(BaseClient):
    """
    hubspot3 Settings client
    Use this to read settings for a given API key, as well as update a setting.
    :see: http://docs.hubapi.com/wiki/Settings_API
    """

    def _get_path(self, subpath: str) -> str:
        """get the full api url for the given subpath on this client"""
        return "settings/{}/{}".format(SETTINGS_API_VERSION, subpath)

    def get_settings(self, **options):
        """Returns the settings we know about for this API key."""
        return self._call("settings", **options)

    def get_setting(self, name: str, **options):
        """Returns the specific requested setting name, if found."""
        params = {"name": name}
        return self._call("settings", params=params, **options)

    def update_setting(self, data: Dict, **options):
        """Updates a specific setting for this API key."""
        params = {}
        if data["name"]:
            params["name"] = data["name"]
        if data["value"]:
            params["value"] = data["value"]

        return self._call(
            "settings", params=params, data=data, method="POST", **options
        )

    def delete_setting(self, name: str, **options):
        """Deletes a specific setting by emptying out its value."""
        params = {}
        if name:
            params["name"] = name
        else:
            raise HubspotError("Setting name required.", "settings")
        return self._call("settings", params=params, method="DELETE", **options)



"""
hubspot ecommerce bridge api
"""
from collections.abc import Mapping, Sequence
from typing import Dict, List

from hubspot3.base import BaseClient
from hubspot3.error import HubspotBadConfig
from hubspot3.utils import get_log

ECOMMERCE_BRIDGE_API_VERSION = "2"
MAX_ECOMMERCE_BRIDGE_SYNC_MESSAGES = 200  # Maximum number of sync messages per request.
MAX_ECOMMERCE_BRIDGE_SYNC_ERRORS = 200  # Maximum number of errors per response.


class EcommerceBridgeClient(BaseClient):
    """
    The hubspot3 Ecommerce Bridge client uses the _make_request method to call the
    API for data.  It returns a python object translated from the json returned
    """

    class ObjectType:
        """object type enum"""

        CONTACT = "CONTACT"
        DEAL = "DEAL"
        PRODUCT = "PRODUCT"
        LINE_ITEM = "LINE_ITEM"

    class Action:
        """action enum"""

        DELETE = "DELETE"
        UPSERT = "UPSERT"

    class ErrorType:
        """error type enum"""

        INACTIVE_PORTAL = "INACTIVE_PORTAL"
        NO_SYNC_SETTINGS = "NO_SYNC_SETTINGS"
        SETTINGS_NOT_ENABLED = "SETTINGS_NOT_ENABLED"
        NO_MAPPINGS_DEFINED = "NO_MAPPINGS_DEFINED"
        MISSING_REQUIRED_PROPERTY = "MISSING_REQUIRED_PROPERTY"
        NO_PROPERTIES_DEFINED = "NO_PROPERTIES_DEFINED"
        INVALID_ASSOCIATION_PROPERTY = "INVALID_ASSOCIATION_PROPERTY"
        INVALID_DEAL_STAGE = "INVALID_DEAL_STAGE"
        INVALID_EMAIL_ADDRESS = "INVALID_EMAIL_ADDRESS"
        INVALID_ENUM_PROPERTY = "INVALID_ENUM_PROPERTY"
        UNKNOWN_ERROR = "UNKNOWN_ERROR"

    def __init__(self, *args, **kwargs):
        """initialize an ecommerce bridge client"""
        super(EcommerceBridgeClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.ecommerce_bridge")

    def _get_path(self, subpath):
        return "extensions/ecomm/v{}/{}".format(ECOMMERCE_BRIDGE_API_VERSION, subpath)

    def send_sync_messages(
        self, object_type: str, messages: Sequence, store_id: str = "default", **options
    ) -> None:
        """
        Send multiple ecommerce sync messages for the given object type and store ID.

        If the number of sync messages exceeds the maximum number of sync messages per request,
        the messages will automatically be split up into appropriately sized requests.

        :see: https://developers.hubspot.com/docs/methods/ecommerce/v2/send-sync-messages
        """
        # Break the messages down into chunks that do not contain more than the maximum number
        # of allowed sync messages per request.
        chunks = [
            list(messages[i : i + MAX_ECOMMERCE_BRIDGE_SYNC_MESSAGES])
            for i in range(0, len(messages), MAX_ECOMMERCE_BRIDGE_SYNC_MESSAGES)
        ]
        for chunk in chunks:
            data = {"objectType": object_type, "storeId": store_id, "messages": chunk}
            self._call("sync/messages", data=data, method="PUT", **options)

    def _get_sync_errors(
        self,
        subpath: str,
        include_resolved: bool = False,
        error_type: str = None,
        object_type: str = None,
        limit: int = None,
        starting_page: int = None,
        **options
    ) -> List:
        """Internal method to retrieve sync errors from an endpoint."""
        # Build the common parameters for all requests.
        query_limit = min(
            limit or MAX_ECOMMERCE_BRIDGE_SYNC_ERRORS, MAX_ECOMMERCE_BRIDGE_SYNC_ERRORS
        )
        common_params = {
            "includeResolved": str(include_resolved).lower(),
            "limit": query_limit,
        }
        if error_type:
            common_params["errorType"] = error_type
        if object_type:
            common_params["objectType"] = object_type

        # Potentially perform multiple requests until the given limit is reached or until all
        # errors have been retrieved.
        errors = []  # type: List
        finished = False
        page = starting_page or 1
        while not finished:
            batch = self._call(
                subpath, method="GET", params=dict(common_params, page=page), **options
            )
            errors.extend(batch["results"])
            # The "paging" attribute is only present if there are more pages.
            finished = "paging" not in batch or (
                limit is not None and len(errors) >= limit
            )
            # The endpoints use some weird kind of pagination where the page parameter is
            # essentially an offset: page 1 starts with record 1, page 2 with record 2, etc.
            # regardless of the limit parameter. Therefore, the next page must be determined by
            # adding the query limit (instead of 1) to get the next set of errors that weren't
            # already part of the current batch.
            page += query_limit

        return errors[:limit]

    def get_sync_errors_for_account(
        self,
        include_resolved: bool = False,
        error_type: str = None,
        object_type: str = None,
        limit: int = None,
        starting_page: int = None,
        **options
    ) -> List:
        """
        Retrieve a list of error dictionaries for an account, optionally filtered/limited, and
        ordered by recency.

        This method and the endpoint it calls can only be used with a portal API key and the portal
        is determined from that key.

        :see: https://developers.hubspot.com/docs/methods/ecommerce/v2/get-all-sync-errors-for-a-specific-account # noqa
        """
        if not self.api_key:
            raise HubspotBadConfig(
                "The app-independent sync errors for a specific account can "
                "only be retrieved using the corresponding portal API key."
            )
        return self._get_sync_errors(
            "sync/errors/portal",
            include_resolved=include_resolved,
            error_type=error_type,
            object_type=object_type,
            limit=limit,
            starting_page=starting_page,
            **options
        )

    def get_sync_errors_for_app(
        self,
        app_id: int,
        include_resolved: bool = False,
        error_type: str = None,
        object_type: str = None,
        limit: int = None,
        starting_page: int = None,
        **options
    ) -> List:
        """
        Retrieve a list of error dictionaries for the app with the given ID, optionally
        filtered/limited, and ordered by recency.

        This method and the endpoint it calls can only be used with the developer API key of the
        developer portal that created the app.

        :see: https://developers.hubspot.com/docs/methods/ecommerce/v2/get-all-sync-errors-for-an-app # noqa
        """
        if not self.api_key:
            raise HubspotBadConfig(
                "The portal-independent sync errors for an app can only be "
                "retrieved using the corresponding developer API key."
            )
        return self._get_sync_errors(
            "sync/errors/app/{app_id}".format(app_id=app_id),
            include_resolved=include_resolved,
            error_type=error_type,
            object_type=object_type,
            limit=limit,
            starting_page=starting_page,
            **options
        )

    def get_sync_errors_for_app_and_account(
        self,
        include_resolved: bool = False,
        error_type: str = None,
        object_type: str = None,
        limit: int = None,
        starting_page: int = None,
        **options
    ):
        """
        Retrieve a list of error dictionaries for an app in specific portal, optionally
        filtered/limited, and ordered by recency.

        This method and the endpoint it calls can only be used with OAuth tokens and both the app
        and the portal are determined from the tokens used.

        :see: https://developers.hubspot.com/docs/methods/ecommerce/v2/get-all-sync-errors-for-an-app-and-account # noqa
        """
        if not self.access_token:
            raise HubspotBadConfig(
                "The sync errors for a specific account from a specific app "
                "can only be retrieved using an access token."
            )
        return self._get_sync_errors(
            "sync/errors",
            include_resolved=include_resolved,
            error_type=error_type,
            object_type=object_type,
            limit=limit,
            starting_page=starting_page,
            **options
        )

    def create_or_update_settings(
        self,
        mappings: Mapping,
        webhook_uri: str = None,
        enabled: bool = True,
        app_id: int = None,
        show_provided_mappings: bool = False,
        **options
    ):
        """
        Create or update the ecommerce settings for a portal or app.
        :see: https://developers.hubspot.com/docs/methods/ecommerce/v2/upsert-settings
        """
        data = {"mappings": dict(mappings), "enabled": enabled}
        if webhook_uri:
            data["webhookUri"] = webhook_uri

        params = {}  # type: Dict
        params["showProvidedMappings"] = str(show_provided_mappings).lower()
        if app_id:
            params["appId"] = app_id

        return self._call("settings", data=data, params=params, method="PUT", **options)

    def create_or_update_store(
        self, store_id: str, label: str, admin_uri: str = None, **options
    ):
        """
        Create or update the store with the given ID.
        :see: https://developers.hubspot.com/docs/methods/ecomm-bridge/v2/create-or-update-store
        """
        data = {"id": store_id, "label": label}
        if admin_uri:
            data["adminUri"] = admin_uri
        return self._call("stores", data=data, method="PUT", **options)

    def check_sync_status_for_object(
        self,
        object_type: str,
        external_object_id: str,
        store_id: str = "default",
        **options
    ) -> Dict:
        """
        Get the synchronization status of a object by its external ID.
        :see: https://developers.hubspot.com/docs/methods/ecommerce/v2/check-sync-status
        """
        return self._call(
            "sync/status/{}/{}/{}".format(store_id, object_type, external_object_id),
            method="GET",
            **options
        )



"""
hubspot cms_templates api client
"""
from hubspot3.base import BaseClient


TEMPLATES_API_VERSION = "2"


class CMSTemplatesClient(BaseClient):
    """
    provides a client for accessing hubspot template info
    """

    def _get_path(self, subpath: str) -> str:
        return "content/api/v{}/templates/{}".format(TEMPLATES_API_VERSION, subpath)

    def get_templates(self, **options):
        return self._call("", **options)

    def get_template_info(self, template_id: str, **options):
        return self._call("{}".format(template_id), **options)

    def get_template_buffer(self, template_id: str, **options):
        return self._call("{}/buffer".format(template_id), **options)

    def get_template_has_buffered_changes(self, template_id: str, **options):
        return self._call("{}/has_buffered_changes".format(template_id), **options)

    def get_template_versions(self, template_id: str, **options):
        return self._call("{}/versions".format(template_id), **options)

    def get_template_version_info(self, template_id: str, version_id: str, **options):
        return self._call("{}/versions/{}".format(template_id, version_id), **options)



"""
hubspot tickets api
"""
from hubspot3.base import BaseClient
from hubspot3.utils import get_log
from typing import Dict, List


TICKETS_API_VERSION = "1"


class TicketsClient(BaseClient):
    """
    hubspot3 Tickets client
    :see: https://developers.hubspot.com/docs/methods/tickets/tickets-overview
    """

    def __init__(self, *args, **kwargs):
        """initialize a tickets client"""
        super(TicketsClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.tickets")

    def _get_path(self, subpath):
        """tickets subpath generator"""
        return "crm-objects/v{}/{}".format(TICKETS_API_VERSION, subpath)

    def create(
        self, pipeline: str, stage: str, properties: Dict = None, **options
    ) -> Dict:
        """
        create a ticket.
        pass in a pipeline and stage, then a key value pair of other properties
        properties will be converted to the name=, value=, format automatically
        :see: https://developers.hubspot.com/docs/methods/tickets/create-ticket
        """
        if not pipeline or not stage:
            raise Exception("pipeline and stage are required to create a ticket!")
        if not properties:
            properties = {}
        ticket_data = [{"name": x, "value": y} for x, y in properties.items()]
        ticket_data.append({"name": "hs_pipeline", "value": pipeline})
        ticket_data.append({"name": "hs_pipeline_stage", "value": stage})
        return self._call("objects/tickets", data=ticket_data, method="POST", **options)

    def update(self, ticket_id: str, data: Dict, **options) -> Dict:
        """
        update a ticket by its ticket id, with the given data
        :see: https://developers.hubspot.com/docs/methods/tickets/update-ticket
        """
        ticket_data = [{"name": x, "value": y} for x, y in data.items()]
        return self._call(
            "objects/tickets/{}".format(ticket_id),
            method="PUT",
            data=ticket_data,
            **options
        )

    def get(
        self,
        ticket_id: str,
        properties: List[str] = None,
        include_deleted: bool = False,
        **options
    ) -> Dict:
        """
        get a ticket by its ticket_id
        :see: https://developers.hubspot.com/docs/methods/tickets/get_ticket_by_id
        """
        properties = properties or [
            "subject",
            "content",
            "hs_pipeline",
            "hs_pipeline_stage",
        ]

        params = options.pop("params", {})
        params.update({"includeDeletes": include_deleted})
        options.update({"params": params})

        return self._call(
            "objects/tickets/{}".format(ticket_id),
            method="GET",
            properties=properties,
            **options
        )

    def get_all(self, properties: List[str] = None, limit: int = -1, **options) -> list:
        """
        Get all tickets in hubspot
        :see: https://developers.hubspot.com/docs/methods/tickets/get-all-tickets
        """
        properties = properties or [
            "subject",
            "content",
            "hs_pipeline",
            "hs_pipeline_stage",
        ]

        finished = False
        output = []  # type: list
        offset = 0
        limited = limit > 0
        while not finished:
            batch = self._call(
                "objects/tickets/paged",
                method="GET",
                params={"offset": offset},
                properties=properties,
                **options
            )
            output.extend(batch["objects"])
            finished = not batch["hasMore"]
            offset = batch["offset"]

        return output if not limited else output[:limit]



"""
hubspot contacts api
"""
import warnings
from typing import Union
from hubspot3.crm_associations import CRMAssociationsClient
from hubspot3.base import BaseClient
from hubspot3.utils import prettify, get_log
from typing import Dict, List


CONTACTS_API_VERSION = "1"


class ContactsClient(BaseClient):
    """
    The hubspot3 Contacts client uses the _make_request method to call the
    API for data.  It returns a python object translated from the json returned
    """

    class Recency:
        """recency type enum"""

        CREATED = "created"
        MODIFIED = "modified"

    def __init__(self, *args, **kwargs):
        """initialize a contacts client"""
        super(ContactsClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.contacts")

    def _get_path(self, subpath: str) -> str:
        """get path"""
        return "contacts/v{}/{}".format(CONTACTS_API_VERSION, subpath)

    def get_by_id(self, contact_id: str, **options):
        """Get contact specified by ID"""
        return self._call(
            "contact/vid/{}/profile".format(contact_id), method="GET", **options
        )

    def get_by_email(self, email: str, **options):
        """Get contact specified by email address."""
        return self._call(
            "contact/email/{email}/profile".format(email=email), method="GET", **options
        )

    def create(self, data: Dict = None, **options):
        """create a contact"""
        data = data or {}
        return self._call("contact", data=data, method="POST", **options)

    def create_or_update_by_email(self, email: str, data: Dict = None, **options):
        """Create or Updates a client with the supplied data."""
        data = data or {}
        return self._call(
            "contact/createOrUpdate/email/{email}".format(email=email),
            data=data,
            method="POST",
            **options
        )

    def update_by_id(self, contact_id: str, data: Dict = None, **options):
        """Update the contact by contact_id with the given data."""
        data = data or {}
        return self._call(
            "contact/vid/{contact_id}/profile".format(contact_id=contact_id),
            data=data,
            method="POST",
            **options
        )

    def update_by_email(self, email: str, data=None, **options):
        """update the concat for the given email address with the given data"""
        data = data or {}

        return self._call(
            "contact/email/{}/profile".format(email),
            data=data,
            method="POST",
            **options
        )

    def delete_by_id(self, contact_id: str, **options):
        """Delete a contact by contact_id."""
        return self._call(
            "contact/vid/{contact_id}".format(contact_id=contact_id),
            method="DELETE",
            **options
        )

    def merge(self, primary_id: int, secondary_id: int, **options):
        """merge the data from the secondary_id into the data of the primary_id"""
        data = dict(vidToMerge=secondary_id)

        self._call(
            "contact/merge-vids/{}/".format(primary_id),
            data=data,
            method="POST",
            **options
        )

    default_batch_properties = [
        "email",
        "firstname",
        "lastname",
        "company",
        "website",
        "phone",
        "address",
        "city",
        "state",
        "zip",
        "associatedcompanyid",
    ]

    def get_batch(self, ids, extra_properties: Union[list, str] = None):
        """given a batch of vids, get more of their info"""
        # default properties to fetch
        properties = set(self.default_batch_properties)

        # append extras if they exist
        if extra_properties:
            if isinstance(extra_properties, list):
                properties.update(extra_properties)
            if isinstance(extra_properties, str):
                properties.add(extra_properties)

        batch = self._call(
            "contact/vids/batch",
            method="GET",
            doseq=True,
            params={"vid": ids, "property": list(properties)},
        )
        # It returns a dict with IDs as keys
        return [prettify(batch[contact], id_key="vid") for contact in batch]

    def link_contact_to_company(self, contact_id, company_id):
        associations_client = CRMAssociationsClient(**self.credentials)
        return associations_client.link_contact_to_company(contact_id, company_id)

    def get_all(
        self,
        extra_properties: Union[list, str] = None,
        limit: int = -1,
        list_id: str = "all",
        **options
    ) -> List[Dict]:
        """
        get all contacts in hubspot, fetching additional properties if passed in
        Can't get phone number from a get-all, so we just grab IDs and
        then have to make ANOTHER call in batches
        :see: https://developers.hubspot.com/docs/methods/contacts/get_contacts
        """
        finished = False
        output = []  # type: list
        offset = 0
        query_limit = 100  # Max value according to docs
        limited = limit > 0
        if limited and limit < query_limit:
            query_limit = limit
        while not finished:
            batch = self._call(
                "lists/{}/contacts/all".format(list_id),
                method="GET",
                params={"count": query_limit, "vidOffset": offset},
                **options
            )
            output.extend(
                self.get_batch(
                    [contact["vid"] for contact in batch["contacts"]],
                    extra_properties=extra_properties,
                )
            )
            finished = not batch["has-more"] or (limited and len(output) >= limit)
            offset = batch["vid-offset"]

        return output if not limited else output[:limit]

    def _get_recent(
        self,
        recency_type: str,
        limit: int = 100,
        vid_offset: int = 0,
        time_offset: int = 0,
        **options
    ) -> List[Dict]:
        """
        return a list of either recently created or recently modified/created contacts
        """
        finished = False
        output = []
        query_limit = 100  # max according to the docs
        recency_string = (
            "all"
            if recency_type == ContactsClient.Recency.CREATED
            else "recently_updated"
        )
        limited = limit > 0
        if limited and limit < query_limit:
            query_limit = limit

        while not finished:
            params = {"count": query_limit}
            if vid_offset and time_offset:
                params["vidOffset"] = vid_offset
                params["timeOffset"] = time_offset
            batch = self._call(
                "lists/{}/contacts/recent".format(recency_string),
                method="GET",
                params=params,
                doseq=True,
                **options
            )
            output.extend([contact for contact in batch["contacts"]])
            finished = not batch["has-more"] or len(output) >= limit
            vid_offset = batch["vid-offset"]
            time_offset = batch["time-offset"]

        return output[:limit]

    def get_recently_created(self, limit: int = 100) -> List[Dict]:
        """
        get recently created contacts
        :see: https://developers.hubspot.com/docs/methods/contacts/get_recently_created_contacts
        """
        return self._get_recent(ContactsClient.Recency.CREATED, limit=limit)

    def get_in_list(
        self,
        list_id: int,
        limit: int = 100,
        vid_offset: int = 0,
        time_offset: int = 0,
        **options
    ) -> List[Dict]:
        """
        return contacts in a list
        """
        finished = False
        output = []
        query_limit = 100  # max according to the docs
        limited = limit > 0
        if limited and limit < query_limit:
            query_limit = limit

        while not finished:
            params = {"count": query_limit}
            if vid_offset and time_offset:
                params["vidOffset"] = vid_offset
                params["timeOffset"] = time_offset
            batch = self._call(
                "lists/{}/contacts/all".format(list_id),
                method="GET",
                params=params,
                doseq=True,
                **options
            )
            output.extend([contact for contact in batch["contacts"]])
            finished = not batch["has-more"] or len(output) >= limit
            vid_offset = batch.get("vid-offset", 0)
            time_offset = batch.get("time-offset", 0)

        return output[:limit]

    def get_recently_modified(self, limit: int = 100) -> List[Dict]:
        """
        get recently modified and created contacts
        :see: https://developers.hubspot.com/docs/methods/contacts/get_recently_updated_contacts
        """
        return self._get_recent(ContactsClient.Recency.MODIFIED, limit=limit)

    def get_contact_by_id(self, contact_id: str, **options):
        warnings.warn(
            "ContactsClient.get_contact_by_id is deprecated in favor of "
            "ContactsClient.get_by_id",
            DeprecationWarning,
        )
        return self.get_by_id(contact_id, **options)

    def get_contact_by_email(self, email: str, **options):
        warnings.warn(
            "ContactsClient.get_contact_by_email is deprecated in favor of "
            "ContactsClient.get_by_email",
            DeprecationWarning,
        )
        return self.get_by_email(email, **options)

    def create_or_update_a_contact(self, email: str, data: Dict = None, **options):
        warnings.warn(
            "ContactsClient.create_or_update_a_contact is deprecated in favor of "
            "ContactsClient.create_or_update_by_email",
            DeprecationWarning,
        )
        return self.create_or_update_by_email(email, data, **options)

    def update(self, contact_id: str, data: Dict = None, **options):
        warnings.warn(
            "ContactsClient.update is deprecated in favor of "
            "ContactsClient.update_by_id",
            DeprecationWarning,
        )
        return self.update_by_id(contact_id, data, **options)

    def update_a_contact(self, contact_id: str, data: Dict = None, **options):
        warnings.warn(
            "ContactsClient.update_a_contact is deprecated in favor of "
            "ContactsClient.update_by_id",
            DeprecationWarning,
        )
        return self.update_by_id(contact_id, data, **options)

    def delete_a_contact(self, contact_id: str, **options):
        warnings.warn(
            "ContactsClient.delete_a_contact is deprecated in favor of "
            "ContactsClient.delete_by_id",
            DeprecationWarning,
        )
        return self.delete_by_id(contact_id, **options)

    def search(self, search_query: str, **options):
        """
        Search among contacts for matches with the given `search_query`.

        Cf: https://developers.hubspot.com/docs/methods/contacts/search_contacts

        Parameters
        ----------
        search_query: str

        Returns
        -------
        list of dict
            The result of the search as a list of contacts.
        """
        finished = False
        offset = 0
        query_limit = 100  # Max value according to docs

        output = []  # type: list

        while not finished:
            batch = self._call(
                "search/query",
                method="GET",
                params={"count": query_limit, "offset": offset, "q": search_query},
                **options
            )

            output += batch["contacts"]

            finished = not batch["has-more"]
            offset = batch["offset"]

        return output

    def delete_all(self):
        """
        Delete all the contacts. Please use it carefully.
        """
        for contact in self.get_all():
            self.delete_by_id(contact["vid"])



"""Command-line interface for the Hubspot client."""
import json
import sys
import types
from fire.core import Fire as fire, _Fire as fire_execute
from fire.helputils import UsageString as build_usage_string
from fire.parser import SeparateFlagArgs as separate_flag_args
from functools import wraps
from hubspot3 import Hubspot3
from hubspot3.base import BaseClient
from hubspot3.leads import LeadsClient
from typing import Callable, Dict, List, Tuple


def get_config_from_file(filename):
    """Return the content of a JSON config file as a dictionary."""
    with open(filename, "r", encoding="utf-8") as file:
        config = json.load(file)
    if not isinstance(config, dict):
        raise RuntimeError(
            'Config file content must be an object, got "{}" instead.'.format(
                type(config).__name__
            )
        )
    return config


class Hubspot3CLIWrapper:
    __doc__ = """
        Hubspot3 CLI

        To get a list of supported operations, call this CLI without the "--help" option.

        The API client can be configured by providing options BEFORE specifying the operation to
        execute. KWARGS are:
        [--config CONFIG_FILE_PATH] {}
    """.format(
        build_usage_string(Hubspot3).split("\n")[-1]
    )

    # Properties to ignore during discovery. The "me" property must be ignored
    # as it would already perform an API request while being discovered and the
    # "usage_limits" property does not contain an API.
    # Extend this tuple if more properties that aren't API clients are added to
    # the Hubspot3 class.
    IGNORED_PROPERTIES = ("me", "usage_limits")

    def __init__(self, **kwargs):
        # If no arguments were supplied at all, the desired outcome is likely
        # the list of operations/clients. Therefore disable authentication to
        # stop the Hubspot3 initializer from raising an exception since there
        # is neither an API key nor an access token.
        if not kwargs:
            kwargs["disable_auth"] = True

        # If a config file was specified, read its settings and merge the CLI
        # options into them.
        config_file = kwargs.pop("config", None)
        if config_file is not None:
            config = get_config_from_file(config_file)
            kwargs = dict(config, **kwargs)

        # Initialize the main client, discover all sub-clients and set them as
        # attributes on this wrapper so Fire can discover them.
        hubspot3 = Hubspot3(**kwargs)
        self._clients = self._discover_clients(hubspot3)
        for attr, client in self._clients.items():
            setattr(self, attr, ClientCLIWrapper(client))

    def __dir__(self):
        return self._clients  # Let Fire only discover the client attributes.

    def __str__(self):
        return "Hubspot3 CLI"

    def _discover_clients(self, hubspot3: Hubspot3) -> Dict[str, BaseClient]:
        """Find all client instance properties on the given Hubspot3 object."""
        clients = {}
        for attr in dir(hubspot3.__class__):
            # Find properties by searching the class first - that way, a call
            # to getattr doesn't run the properties code on the object.
            if (
                attr.startswith("_")
                or attr in self.IGNORED_PROPERTIES
                or not isinstance(getattr(hubspot3.__class__, attr), property)
            ):
                continue
            client = getattr(hubspot3, attr)
            if isinstance(client, BaseClient):
                clients[attr] = client
        return clients


class ClientCLIWrapper:
    __doc__ = Hubspot3CLIWrapper.__doc__

    # Mapping (client class to attribute names) to define methods that should
    # be ignored during method discovery.
    # Extend this mapping if more methods that aren't API methods are added to
    # a client class.
    IGNORED_METHODS = {LeadsClient: ("camelcase_search_options",)}  # type: Dict
    STDIN_TOKEN = "__stdin__"  # Argument value to trigger stdin parsing.

    def __init__(self, client: BaseClient) -> None:
        self._client_name = client.__class__.__name__
        # Discover all API methods and set them as attributes on this wrapper
        # so Fire can discover them.
        self._methods = self._discover_methods(client)
        for attr, method in self._methods.items():
            setattr(self, attr, self._build_method_wrapper(method))

    def __dir__(self):
        return self._methods  # Let Fire only discover the API methods.

    def __str__(self):
        return "Hubspot3 {} CLI".format(self._client_name)

    def _discover_methods(self, client: BaseClient) -> Dict[str, types.MethodType]:
        """Find all API methods on the given client object."""
        methods = {}
        for attr in dir(client):
            if attr.startswith("_") or attr in self.IGNORED_METHODS.get(
                client.__class__, ()
            ):
                continue
            method = getattr(client, attr)
            if isinstance(method, types.MethodType):
                methods[attr] = method
        return methods

    def _build_method_wrapper(self, method: types.MethodType) -> Callable:
        """Build a wrapper function around the given API method."""

        @wraps(method)
        def wrapper(*args, **kwargs):
            # Replace the stdin token with the actual stdin value and call the
            # original API method.
            args, kwargs = self._replace_stdin_token(*args, **kwargs)
            result = method(*args, **kwargs)

            # Try to ensure to always write JSON to stdout, but don't hide any
            # result either if it can't be JSON-encoded.
            if isinstance(result, bytes):
                result = result.decode("utf-8")
            try:
                result = json.dumps(result)
            except Exception:
                pass
            print(result)

        wrapper.__doc__ = self._build_wrapper_doc(method)
        return wrapper

    def _build_wrapper_doc(self, method: types.MethodType) -> str:
        """Build a helpful docstring for a wrapped API method."""
        return "\n".join(
            (
                method.__doc__ or "",
                "",
                "Supported ARGS/KWARGS are:",
                build_usage_string(method),
                "",
                'The token "{}" may be used as an argument value, which will cause JSON data to be '
                "read from stdin and used as the actual argument value.".format(
                    self.STDIN_TOKEN
                ),
            )
        )

    def _replace_stdin_token(self, *args_tuple, **kwargs) -> Tuple[List, Dict]:
        """
        Replace the values of all given arguments with the JSON-parsed value
        from stdin if their current value is the STDIN_TOKEN.
        """
        args = list(args_tuple)
        stdin_indices = [
            index for index, value in enumerate(args) if value == self.STDIN_TOKEN
        ]
        stdin_keys = [key for key, value in kwargs.items() if value == self.STDIN_TOKEN]
        if stdin_indices or stdin_keys:
            value = json.load(sys.stdin)
            for index in stdin_indices:
                args[index] = value
            for key in stdin_keys:
                kwargs[key] = value
        return args, kwargs


def split_args() -> Tuple[List, List, List]:
    """
    Split system args into three group of argument lists.

    This method will separate all system args in client, API call and fire args
    so that is much easier to instantiate them independently with their own
    arguments.
    """
    args = sys.argv[1:]
    if args == ["--help"]:
        # If the user only called the CLI with "--help", pass it through as an
        # argument for Fire to invoke its help functionality.
        return [], [], args

    # Use a Fire function to split away the Fire arguments.
    args, fire_args = separate_flag_args(args)

    # Search for an argument that represents the sub-client that should be
    # used. Since the CLI arguments have a fixed pattern (client options,
    # sub-client name, method name, method options), this must be the first
    # argument that isn't a named client option argument.
    api_index = 0
    while api_index < len(args):
        arg = args[api_index]
        # Named options can be passed as "--key=value" or "--key value", so the
        # next argument to look at is either the next argument or the one after
        # that, respectively.
        if arg.startswith("--"):
            if "=" not in arg:
                api_index += 1
            api_index += 1
        else:
            break
    client_args, call_args = args[:api_index], args[api_index:]
    return client_args, call_args, fire_args


def main():
    """
    Main routine.

    Split the execution up into two Fire calls: one that creates the main
    Hubspot3CLIWrapper instance and one that works on this instance. That way,
    the argument separator is not required and sensible details like the API
    key are not printed out.
    """
    client_args, call_args, fire_args = split_args()

    if fire_args:
        # If there are arguments for an actual API method call, append the Fire
        # arguments to that second call. Otherwise, append them to the first
        # call as there won't be a second one.
        (call_args or client_args).extend(["--"] + fire_args)
    if call_args:
        # Use the non-printing Fire routine for the first call as only the
        # result of the second, actual API call should be printed.
        component_trace = fire_execute(Hubspot3CLIWrapper, client_args, {}, __package__)
        wrapper = component_trace.GetResult()
        fire(wrapper, call_args)
    else:
        fire(Hubspot3CLIWrapper, client_args, __package__)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
pip setup file
"""
import os
import re
from setuptools import setup


with open("README.md") as readme:
    LONG_DESCRIPTION = readme.read()


def find_version(*file_paths):
    """
    This pattern was modeled on a method from the Python Packaging User Guide:
        https://packaging.python.org/en/latest/single_source_version.html
    We read instead of importing so we don't get import errors if our code
    imports from dependencies listed in install_requires.
    """
    base_module_file = os.path.join(*file_paths)
    with open(base_module_file) as module_file:
        base_module_data = module_file.read()
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", base_module_data, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="hubspot3",
    version=find_version("hubspot3", "globals.py"),
    description=(
        "A python wrapper around HubSpot's APIs, for python 3. "
        "Built initially around hapipy, but heavily modified."
    ),
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="HubSpot Dev Team, Jacobi Petrucciani",
    author_email="jacobi@mimirhq.com",
    url="https://github.com/jpetrucciani/hubspot3.git",
    download_url="https://github.com/jpetrucciani/hubspot3.git",
    license="LICENSE",
    packages=["hubspot3"],
    install_requires=["typing_extensions"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    zip_safe=False,
    extras_require={"cli": ["fire==0.1.3"]},
    entry_points={"console_scripts": ["hubspot3=hubspot3.__main__:main"]},
)



"""
hubspot owners api
"""
from hubspot3.crm_associations import CRMAssociationsClient
from hubspot3.base import BaseClient


OWNERS_API_VERSION = "v2"


class OwnersClient(BaseClient):
    """
    hubspot3 Owners client
    :see: https://developers.hubspot.com/docs/methods/owners/owners_overview
    """

    def _get_path(self, subpath):
        """get the full api url for the given subpath on this client"""
        return "owners/{}/owners".format(OWNERS_API_VERSION)

    def get_owners(self, **options):
        """Only returns the list of owners, does not include additional metadata"""
        return self._call("owners", **options)

    def get_owner_name_by_id(self, owner_id: str, **options) -> str:
        """Given an id of an owner, return their name"""
        owner_name = "value_missing"
        owners = self.get_owners(**options)
        for owner in owners:
            if int(owner["ownerId"]) == int(owner_id):
                owner_name = "{} {}".format(owner["firstName"], owner["lastName"])
        return owner_name

    def get_owner_email_by_id(self, owner_id: str, **options) -> str:
        """given an id of an owner, return their email"""
        owner_email = "value_missing"
        owner = self.get_owner_by_id(owner_id, **options)
        if owner:
            owner_email = owner["email"]
        return owner_email

    def get_owner_by_id(self, owner_id, **options):
        """Retrieve an owner by its id."""
        owners = self.get_owners(**options)
        for owner in owners:
            if int(owner["ownerId"]) == int(owner_id):
                return owner
        return None

    def get_owner_by_email(self, owner_email: str, **options):
        """
        Retrieve an owner by its email.
        """
        owners = self.get_owners(method="GET", params={"email": owner_email}, **options)
        if owners:
            return owners[0]
        return None

    def link_owner_to_company(self, owner_id, company_id):
        """
        Link an owner to a company by using their ids.
        """
        associations_client = CRMAssociationsClient(**self.credentials)
        return associations_client.link_owner_to_company(owner_id, company_id)



"""
hubspot lines api
"""
from typing import Dict, Union
from hubspot3.base import BaseClient
from hubspot3.crm_associations import CRMAssociationsClient
from hubspot3.utils import get_log, prettify, ordered_dict


LINES_API_VERSION = "1"


class LinesClient(BaseClient):
    """
    Line Items API endpoint
    :see: https://developers.hubspot.com/docs/methods/line-items/line-items-overview
    """

    def __init__(self, *args, **kwargs):
        super(LinesClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.lines")

    def _get_path(self, subpath: str):
        return "crm-objects/v{}/objects/line_items/{}".format(
            self.options.get("version") or LINES_API_VERSION, subpath
        )

    def create(self, data=None, **options) -> Dict:
        """
        Create a line item.
        :see: https://developers.hubspot.com/docs/methods/line-items/create-line-item
        """
        return self._call("", data=data, method="POST", **options)

    def delete(self, line_id: int, **options) -> Dict:
        """
        Delete a line item by its ID.
        :see: https://developers.hubspot.com/docs/methods/line-items/delete-line-item
        """
        return self._call("{}".format(line_id), method="DELETE", **options)

    def get(self, line_id: int, **options) -> Dict:
        """
        Retrieve a line by its ID.
        :see: https://developers.hubspot.com/docs/methods/line-items/get_line_item_by_id
        """
        return self._call("{}".format(line_id), **options)

    def update(self, line_id: int, data=None, **options) -> Dict:
        """
        Update an existing line by its ID.
        :see: https://developers.hubspot.com/docs/methods/line-items/update-line-item
        """
        data = data or {}
        return self._call("{}".format(line_id), data=data, method="PUT", **options)

    def get_all(
        self,
        offset: int = 0,
        extra_properties: Union[list, str] = None,
        limit: int = -1,
        **options
    ):
        """
        Retrieve all the line items in the Hubspot account.

        Cf: https://developers.hubspot.com/docs/methods/line-items/get-all-line-items

        :param offset Used to get the next set of results.
        :param extra_properties By default, only the ID, the hubspot product id (`hs_product_id`)
        a few other system fields are returned for the line items. This method with also ask for
        basic properties such as the 'name', the 'price' and the 'quantity'. More could be
        retrieved by using 'extra_properties'.
        :param limit: could be used to prevent to fetch the entire results. Default value is `-1`,
        meaning unlimited.
        """
        finished = False
        output = []
        limited = limit > 0

        # Default properties to fetch
        properties = ["name", "price", "quantity"]

        # append extras if they exist
        if extra_properties:
            if isinstance(extra_properties, list):
                properties += extra_properties
            if isinstance(extra_properties, str):
                properties.append(extra_properties)

        while not finished:
            batch = self._call(
                "paged",
                method="GET",
                params=ordered_dict({"offset": offset, "properties": properties}),
                doseq=True,
                **options
            )
            output.extend(
                [
                    prettify(line_item, id_key="objectId")
                    for line_item in batch["objects"]
                    if not line_item["isDeleted"]
                ]
            )
            finished = not batch["hasMore"] or (limited and len(output) >= limit)
            offset = batch["offset"]

        return output if not limited else output[:limit]

    def link_line_item_to_deal(self, line_item_id, deal_id) -> Dict:
        """Link a line item to a deal."""
        associations_client = CRMAssociationsClient(**self.credentials)
        return associations_client.link_line_item_to_deal(line_item_id, deal_id)



"""
hubspot crm_associations api
"""
from hubspot3.base import BaseClient
from hubspot3.utils import get_log
from enum import Enum
from typing import Union


ASSOCIATIONS_API_VERSION = "1"


class Definitions(Enum):
    """
    :see: https://developers.hubspot.com/docs/methods/crm-associations/crm-associations-overview
    """

    CONTACT_TO_COMPANY = 1
    COMPANY_TO_CONTACT = 2
    DEAL_TO_CONTACT = 3
    CONTACT_TO_DEAL = 4
    DEAL_TO_COMPANY = 5
    COMPANY_TO_DEAL = 6
    COMPANY_TO_ENGAGEMENT = 7
    ENGAGEMENT_TO_COMPANY = 8
    CONTACT_TO_ENGAGEMENT = 9
    ENGAGEMENT_TO_CONTACT = 10
    DEAL_TO_ENGAGEMENT = 11
    ENGAGEMENT_TO_DEAL = 12
    PARENT_COMPANY_TO_CHILD_COMPANY = 13
    CHILD_COMPANY_TO_PARENT_COMPANY = 14
    CONTACT_TO_TICKET = 15
    TICKET_TO_CONTACT = 16
    TICKET_TO_ENGAGEMENT = 17
    ENGAGEMENT_TO_TICKET = 18
    DEAL_TO_LINE_ITEM = 19
    LINE_ITEM_TO_DEAL = 20
    COMPANY_TO_TICKET = 25
    TICKET_TO_COMPANY = 26
    DEAL_TO_TICKET = 27
    TICKET_TO_DEAL = 28
    OWNER_TO_COMPANY = 41


class CRMAssociationsClient(BaseClient):
    """
    Associations extension for Associations API endpoint
    :see: https://developers.hubspot.com/docs/methods/crm-associations/crm-associations-overview
    """

    def __init__(self, *args, **kwargs):
        super(CRMAssociationsClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.crm_associations")

    def _get_path(self, subpath: str) -> str:
        return "crm-associations/v{}/{}".format(
            self.options.get("version") or ASSOCIATIONS_API_VERSION, subpath
        )

    def get(self, object_id: str, definition: Union[Definitions, int]):
        """
        get all associations for the defined object
        :param object_id: Object ID for the object you're looking up
        :param definition: Definition ID for the objects you're looking for associations of
        """
        finished = False
        output = []
        offset = 0
        query_limit = 100  # Max value according to docs
        definition_id = definition if isinstance(definition, int) else definition.value

        while not finished:
            batch = self._call(
                "associations/{}/{}/{}".format(
                    object_id, "HUBSPOT_DEFINED", definition_id
                ),
                method="GET",
                params={"limit": query_limit, "offset": offset},
            )
            output.extend([id_ for id_ in batch["results"]])
            finished = not batch["hasMore"]
            offset = batch["offset"]

        return output

    def get_all(self, object_id: str, definition: Union[Definitions, int], **options):
        """
        get all crm associations
        :param object_id: Object ID for the object you're looking up
        :param definition: Definition ID for the objects you're looking for associations of
        """
        finished = False
        output = []
        offset = 0
        query_limit = 100  # Max value according to docs
        definition_id = definition if isinstance(definition, int) else definition.value

        while not finished:
            batch = self._call(
                "associations/{}/{}/{}".format(
                    object_id, "HUBSPOT_DEFINED", definition_id
                ),
                method="GET",
                params={"limit": query_limit, "offset": offset},
            )
            output.extend([id_ for id_ in batch["results"]])
            finished = not batch["hasMore"]
            offset = batch["offset"]

        return output

    def create(
        self,
        from_object: str,
        to_object: str,
        definition: Union[Definitions, int],
        **options
    ):
        """
        create a hubspot association
        :param from_object: ID of object to relate
        :param to_object: ID of object to relate to
        :param definition: Definition ID for the objects you're looking for associations of
        """
        definition_id = definition if isinstance(definition, int) else definition.value

        return self._call(
            "associations",
            method="PUT",
            data={
                "fromObjectId": from_object,
                "toObjectId": to_object,
                "category": "HUBSPOT_DEFINED",
                "definitionId": definition_id,
            },
            **options
        )

    def delete(
        self,
        from_object: str,
        to_object: str,
        definition: Union[Definitions, int],
        **options
    ):
        """
        delete a hubspot association
        :param from_object: ID of object to relate
        :param to_object: ID of object to relate to
        :param definition: Definition ID for the objects you're looking for associations of
        """
        definition_id = definition if isinstance(definition, int) else definition.value

        return self._call(
            "associations/delete",
            method="PUT",
            data={
                "fromObjectId": from_object,
                "toObjectId": to_object,
                "category": "HUBSPOT_DEFINED",
                "definitionId": definition_id,
            },
            **options
        )

    def get_deal_to_lines_items(self, deal_id: str):
        """Get the lines related to a deal."""
        return self.get(object_id=deal_id, definition=Definitions.DEAL_TO_LINE_ITEM)

    def get_company_to_contacts(self, company_id: str):
        """Get the contacts related to a company."""
        return self.get(object_id=company_id, definition=Definitions.COMPANY_TO_CONTACT)

    def get_company_to_deals(self, company_id: str):
        """Get the deals related to a company."""
        return self.get(object_id=company_id, definition=Definitions.COMPANY_TO_DEAL)

    def link_line_item_to_deal(self, line_item_id: str, deal_id: str):
        """Create a new association between a line item and a deal."""
        return self.create(
            from_object=line_item_id,
            to_object=deal_id,
            definition=Definitions.LINE_ITEM_TO_DEAL,
        )

    def link_contact_to_company(self, contact_id: str, company_id: str):
        """Create a new association between a contact and a company."""
        return self.create(
            from_object=contact_id,
            to_object=company_id,
            definition=Definitions.CONTACT_TO_COMPANY,
        )

    def link_owner_to_company(self, owner_id: str, company_id: str):
        """Create a new association between an owner and a company."""
        return self.create(
            from_object=owner_id,
            to_object=company_id,
            definition=Definitions.OWNER_TO_COMPANY,
        )
